"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const python_bridge_1 = require("python-bridge");
const config_1 = require("../config");
class PythonFacenet {
    constructor() {
        this.facenetInited = false;
        this.mtcnnInited = false;
        this.SRC_DIRNAME = __dirname;
        // python -m venv $VENV: the directory of vent
        this.VENV = path.join(config_1.MODULE_ROOT, 'python3');
        config_1.log.verbose('PythonFacenet', 'constructor() SRC=%s', this.SRC_DIRNAME);
        this.initVenv();
        this.python3 = this.initBridge();
    }
    initVenv() {
        config_1.log.verbose('PythonFacenet', `initVenv() to ${this.VENV}`);
        const PATH = `${this.VENV}/bin:` + process.env['PATH'];
        Object.assign(process.env, {
            VIRTUAL_ENV: this.VENV,
            PATH,
        });
        delete process.env['PYTHONHOME'];
    }
    initBridge() {
        config_1.log.verbose('PythonFacenet', 'initBridge()');
        const TF_CPP_MIN_LOG_LEVEL = '2'; // suppress tensorflow warnings
        let PYTHONPATH = [
            `${this.VENV}/facenet/src`,
            this.SRC_DIRNAME,
        ].join(':');
        if (process.env['PYTHONPATH']) {
            PYTHONPATH += ':' + process.env['PYTHONPATH'];
        }
        Object.assign(process.env, {
            PYTHONPATH,
            TF_CPP_MIN_LOG_LEVEL,
        });
        const bridge = python_bridge_1.pythonBridge({
            python: 'python3',
        });
        return bridge;
    }
    initFacenet() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PythonFacenet', 'initFacenet()');
            if (this.facenetInited) {
                return;
            }
            const start = Date.now();
            yield this.python3.ex `
      from facenet_bridge import FacenetBridge
      facenet_bridge = FacenetBridge()
      facenet_bridge.init()
    `;
            config_1.log.silly('PythonFacenet', 'initFacenet() facenet_bridge.init() cost %d milliseconds', Date.now() - start);
            this.facenetInited = true;
        });
    }
    initMtcnn() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PythonFacenet', 'initMtcnn()');
            if (this.mtcnnInited) {
                return;
            }
            const start = Date.now();
            // we need not to care about session.close()(?)
            yield this.python3.ex `
      from facenet_bridge import MtcnnBridge
      mtcnn_bridge = MtcnnBridge()
      mtcnn_bridge.init()
    `;
            config_1.log.silly('PythonFacenet', 'initMtcnn() mtcnn_bridge.init() cost milliseconds.', Date.now() - start);
            this.mtcnnInited = true;
        });
    }
    quit() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PythonFacenet', 'quit()');
            if (!this.python3) {
                throw new Error('no phthon3 bridge inited yet!');
            }
            yield this.python3.end();
            this.mtcnnInited = this.facenetInited = false;
        });
    }
    /**
     *
     * @param imageData
     */
    align(imageData) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PythonFacenet', 'align(%dx%d)', imageData.width, imageData.height);
            yield this.initMtcnn();
            const row = imageData.height;
            const col = imageData.width;
            const depth = imageData.data.length / row / col;
            const base64Text = this.base64ImageData(imageData);
            let boundingBoxes;
            let landmarks;
            const start = Date.now();
            [boundingBoxes, landmarks] = yield this.python3 `mtcnn_bridge.align(${base64Text}, ${row}, ${col}, ${depth})`;
            config_1.log.silly('PythonFacenet', 'align() mtcnn_bridge.align() cost %d milliseconds', Date.now() - start);
            return [boundingBoxes, landmarks];
        });
    }
    /**
     *
     * @param image
     */
    embedding(imageData) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PythonFacenet', 'embedding(%dx%d)', imageData.width, imageData.height);
            if (!this.facenetInited) {
                yield this.initFacenet();
            }
            const row = imageData.height;
            const col = imageData.width;
            const depth = imageData.data.length / row / col;
            const base64Text = this.base64ImageData(imageData);
            const start = Date.now();
            const embedding = yield this.python3 `facenet_bridge.embedding(${base64Text}, ${row}, ${col}, ${depth})`;
            config_1.log.silly('PythonFacenet', 'embedding() facenet_bridge.embedding() cost %d milliseconds', Date.now() - start);
            return embedding;
        });
    }
    // public async json_parse(text: string): Promise<any> {
    //   await this.initPythonBridge()
    //   await this.python.ex`from facenet_bridge import json_parse`
    //   return await this.python`json_parse(${text})`
    // }
    base64_to_image(text, row, col, depth) {
        return __awaiter(this, void 0, void 0, function* () {
            // await this.initPythonBridge()
            yield this.python3.ex `from facenet_bridge import base64_to_image`;
            return yield this.python3 `base64_to_image(${text}, ${row}, ${col}, ${depth}).tolist()`;
        });
    }
    /**
     * Deal with big file(e.g. 4000 x 4000 JPEG)
     * the following method will cause NODEJS HEAP MEMORY OUT(>1.5GB)
     * https://github.com/nicolaspanel/numjs/issues/21#issuecomment-319301957
     *
     * MEMORY OUT 1: image.flatten()
     * MEMORY OUT 2: [].concat.apply([], arrays);
     * MEMORY OUT 3: image.reshape()
     *
     * @param image
     */
    base64ImageData(imageData) {
        return Buffer.from(imageData.data.buffer)
            .toString('base64');
        // const [row, col, depth] = image.shape
        // const typedData = new Uint8ClampedArray(row * col * depth)
        // let n = 0
        // for (let i = 0; i < row; i++) {
        //   for (let j = 0; j < col; j++) {
        //     for (let k = 0; k < depth; k++) {
        //       typedData[n++] = image.get(i, j, k) as any as number
        //     }
        //   }
        // }
        // const base64Text = Buffer.from(typedData.buffer)
        //                         .toString('base64')
        // return base64Text
    }
}
exports.PythonFacenet = PythonFacenet;
//# sourceMappingURL=python-facenet.js.map